/*
 * arena.h
 *
 * Arena and ArenaConstIterator
 *
 * Ulrik Petersen
 * Created: 3/2-2005
 * Last update: 7/15-2008
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2008  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file arena.h
 *@brief Header file for Arena and related classes (EMdF layer)
 */


#ifndef ARENA__H__
#define ARENA__H__

#include <cstring>
#include <string>
#include <cstring>
#include <cstdlib>
#include <debug.h>

/**
 *\defgroup Arena Arena: Chunks of memory.
 *\ingroup EMdF
 */



class ArenaConstIterator; // forward declaration
class Bigstring; // forward declaration

/**\internal
 * @brief Arena: Chunks of easily allocated, easily freed memory.
 *\ingroup Arena
 * 
 * An "Arena", i.e., a linked listof large chunks of memory from which
 * one can allocate memory easily and then delete it all in one go.
 */
class Arena {
	friend class ArenaConstIterator;
 protected:
	/** A chunk memory.
	 *\ingroup Arena
	 *
	 * Part of the the linked list of chunks in an Arena.
	 */
	class Chunk {
	private:
		friend class Arena;
		friend class ArenaConstIterator;
		friend class Bigstring;
		enum { chunk_size = 512 * 1024 /**< The chunk size */
		}; // 512KB
		Chunk *m_next; /**< Pointer to the next element in the linked list. */
		int m_index; /**< Index of the first free byte in m_bytes. */
		char m_bytes[chunk_size]; /**< The bytes of the chunk. */
		/** Constructor: Constructs a fully free chunk with m_next=0. */
		Chunk() : m_next(0), m_index(0) {};
		/** Constructor: Constructs a fully free chunk with m_next=pNext. 
		 * @param pNext The next Chunk. Afterwards. m_next == pNext. 
		 */
		Chunk(Chunk *pNext) : m_next(pNext), m_index(0) {};
		/** Destructor. Also deletes m_next. */
		~Chunk() { delete m_next; }
		/** Set m_next.
		 * @param pNext The value to set m_next to.
		 */
		void setNext(Chunk *pNext) { m_next = pNext; };
	};
	Chunk *m_tail; /**< The tail of the linked list. */
	Chunk *m_head; /**< The head of the linked list. */
	
 public:
	/** Constructor. Initializes to one chunk.*/
	Arena() { m_head = m_tail = new Chunk(); };
	/** Destructor. Destroys all chunks. */
	~Arena() { delete m_head; };
	
	/** Allocate a piece of memory of the given size.
	 *
	 * If the current chunk is not able to hold the given size,
	 * grow() is called.
	 *
	 * NOTE that size must be strictly less than Chunk::chunk_size.
	 *
	 * @param size The size of the piece of memory to allocate.
	 *
	 * @return A pointer within some Chunk to a piece of memory which is
	 * now allocated.
	 */
	void *allocate(int size) {
		if (m_tail->m_index + size >= Chunk::chunk_size) {
			grow();
		}
		ASSERT_THROW(m_tail->m_index + size < Chunk::chunk_size,
			     "Chunk not grown correctly");
		int current_index = m_tail->m_index;
		m_tail->m_index += size;
		return (void *) &(m_tail->m_bytes[current_index]);
	};
	ArenaConstIterator const_iterator(unsigned int size) const;
 private:
	/** Add a new chunk.*/
	void grow(void) { 
		Chunk *pCurrentTail = m_tail; 
		m_tail = new Chunk(0); 
		pCurrentTail->setNext(m_tail); 
	};
};



/** Const iterator on specific size chunks in an Arena.
 *\ingroup Arena
 *
 * This iterates over the contents of chunks in an Arena.
 *
 * Each content must have a specific size.
 *
 * The first content must be at the beginning of the first chunk.
 */
class ArenaConstIterator {
 private:
	Arena::Chunk *m_current;
	int m_index;
	const Arena *m_pMotherArena;
	unsigned int m_size;
 public:
	ArenaConstIterator();
	ArenaConstIterator(unsigned int size, const Arena *pMotherArena);
	ArenaConstIterator(const ArenaConstIterator& other);
	~ArenaConstIterator();
	bool hasNext() const; // Is the iterator == end iterator?  Doesn't alter iterator
	void *next(); // Gets current and advances iterator afterwards
	void *current(); // Gets current without altering iterator
	ArenaConstIterator& operator=(const ArenaConstIterator& other);
 private:
	void assign(const ArenaConstIterator& other);
};


/** A "big string".
 *\ingroup Arena
 *
 * This is a string which is capable of holding lots and lots of data,
 * even many megabytes.
 *
 * This is useful on Windows Visual C++, where the implementation of
 * std::string chokes on strings larger than about 1MB.
 */
class Bigstring : protected Arena {
 public:
	/** Constructor. Constructs an empty string. */
	Bigstring() : Arena() {};
	/* Destructor. Deallocates all memory. */
	~Bigstring() {};
	/** Add a single char.
	 *
	 * @param c the char to add.
	 */
	void addChar(char c) { void *pMem = allocate(1); *((char *)pMem)  = c; };
	/** Add a zero-delimited string.
	 * 
	 * @param sz The zero-delimited string to add. The trailing
	 * zero is not added.  */
	void addsz(const char *sz) { size_t len = strlen(sz); void *pMem = allocate(len); strcpy((char*) pMem, sz); };
	/** Convert to a std::string.
	 *
	 * @param str The output std::string to which to "export".
	 */
	void toString(std::string& str) {
		str = "";
		Chunk *pChunk = m_head;
		while (pChunk->m_next != 0) {
			str.append((char*) pChunk->m_bytes, pChunk->m_index);
			pChunk = pChunk->m_next;
		}
		if (pChunk->m_index > 0)  {
			str.append((char*) pChunk->m_bytes, pChunk->m_index);
		}
	};
	/** Test if we are empty. 
	 * @return true iff the string is the empty string. 
	 */
	bool empty() const { return !const_iterator(1).hasNext(); };
};




#endif // ARENA__H__
