/*
 * json.h
 *
 * JSON classes
 *
 * Ulrik Sandborg-Petersen
 * Created: 7/28-2008
 * Last update: 8/19-2008
 *
 */

#ifndef JSON_HEADER_H_
#define JSON_HEADER_H_

#include <string>
#include <map>
#include <list>
#include <ostream>
#include <json_lexer.h>

typedef enum {
	kJSONNull,
	kJSONBoolean,
	kJSONInteger,
	kJSONString,
	kJSONList,
	kJSONObject
} eJSONValueKind;

class JSONValue; // Forward declaration

class JSONKeyValuePair {
 protected:
	std::string *m_pKey;
	JSONValue *m_pValue;
	JSONKeyValuePair *m_pNext;
 public:
	JSONKeyValuePair(std::string *pKey, JSONValue *pValue, JSONKeyValuePair *pNext);
	~JSONKeyValuePair();

	std::string getKey(void) const { return *m_pKey; };
	JSONValue *extractValue(void) { JSONValue *pResult = m_pValue; m_pValue = 0; return pResult; };

	void setNext(JSONKeyValuePair *pNext) { m_pNext = pNext; };
	JSONKeyValuePair *getNext(void) { return m_pNext; };
 private:
	JSONKeyValuePair(const JSONKeyValuePair& other);
	JSONKeyValuePair& operator=(const JSONKeyValuePair& other);
};

class JSONListElement {
 protected:
	JSONValue *m_pValue;
	JSONListElement *m_pNext;
 public:
	JSONListElement(JSONValue *pValue);
	~JSONListElement();
	JSONValue *extractValue(void) { JSONValue *pResult = m_pValue; m_pValue = 0; return pResult; };
	void setNext(JSONListElement *pNext) { m_pNext = pNext; };
	JSONListElement *getNext(void) { return m_pNext; };
 private:
	JSONListElement(const JSONListElement& other);
	JSONListElement& operator=(const JSONListElement& other);
};

class JSONValue {
	eJSONValueKind m_kind;
	union {
		bool m_boolean;
		long m_integer;
		const std::string* m_pString;
		std::list<const JSONValue*> *m_pList;
		std::map<std::string, const JSONValue*> *m_pObject;
	} m_value;
 public:
	JSONValue();                        // null
	JSONValue(bool bValue);             // boolean
	JSONValue(long integer);            // integer
	JSONValue(std::string *pString);    // string
	JSONValue(JSONListElement *pTail);  // list
	JSONValue(JSONKeyValuePair *pTail); // object
	~JSONValue();
	eJSONValueKind getKind(void) const { return m_kind; };
	std::string getString(void) const;
	bool getBoolean(void) const;
	long getInteger(void) const;
	const std::list<const JSONValue*>& getList(void) const;
	const JSONValue* getObjectValue(const std::string& key) const;
	const JSONValue* operator[](const std::string& key) const { return getObjectValue(key); };
	std::list<std::string> getObjectKeys(void) const;
	bool hasObjectKey(const std::string& key) const;
	void pretty(std::ostream *pOut, int indent_level = 0) const;
 private:
	JSONValue(const JSONValue& other);
	JSONValue& operator=(const JSONValue& other);
	void deleteList(void);
	void deleteObject(void);
	void indent(std::ostream *pOut, int indent_level) const;
};

class JSONExecEnv {
 protected:
	std::string *m_pError;
	JSONValue *m_pTopValue;
 public:
	JSONScanner *pLexer;
	JSONExecEnv();
	~JSONExecEnv();
	bool bSyntaxError;

	void setTopValue(JSONValue *pValue);
	JSONValue *extractTopValue(void);

	void clearError(void);
	std::string getError();
	void addToError(const std::string& addition);
	void prependError(const std::string& addition);

	void clean();
};


extern JSONValue *parseJSON(JSONExecEnv* pEE, const std::string& json_string, bool& bResult);

#endif
