/*
 * renderobjects.h
 *
 * Classes and methods to render Emdros objects into something else,
 * based on a JSON template.
 *
 * Ulrik Sandborg-Petersen
 * Created: 4/22-2007
 * Last update: 9/15-2008
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2007-2008  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/


#ifndef RENDEROBJECTS__H__
#define RENDEROBJECTS__H__

#include <map>
#include <vector>
#include <string>
#include <list>

#include <environment_emdros.h>
#include <monads.h>
#include <memobject.h>
#include <json_classes.h>
#include <templatelang_classes.h>


class RenderObjects {
 private:
	typedef std::list<std::string> ListOfString;
	typedef std::list<id_d_t> ListOfID_D;
	typedef std::map<std::string, int> OTN2IntMap;
	typedef std::map<monad_m, ListOfID_D> Monad2ID_DListMap;
	typedef std::vector<ListOfID_D> Monad2ID_DListVector;
	typedef std::vector<MemObject*> Monad2MemObjectVector;
	typedef std::map<id_d_t, MemObject*> ID_D2MemObjectMap;
	typedef std::map<std::string, std::string*> String2PStringMap;
	typedef std::map<std::string, TemplateASTNode*> String2TemplateASTNodeMap;

	ListOfString m_xml;
	EmdrosEnv *m_pEE;
	const JSONValue *m_pFetchInfo;
	const JSONValue *m_pStyleSheet;
	const JSONValue *m_pObjectTypes;

	std::string m_stylesheet;
	
	TemplateLangExecEnv *m_pTemplateEE;

	monad_m m_fm, m_lm;

	String2PStringMap m_OTNPs;
	ListOfString m_OTHandlingPriorityList;
	String2TemplateASTNodeMap m_startTemplates;
	String2TemplateASTNodeMap m_endTemplates;
	OTN2IntMap m_startPriority;
	OTN2IntMap m_endPriority;
	ID_D2MemObjectMap m_objs;
	Monad2ID_DListMap m_fms, m_lms;
	std::string m_db_name;
	std::string m_error_message;
	bool m_bIsOK;
 public:
	RenderObjects(EmdrosEnv *pEnv, const std::string& db_name, const JSONValue *pFetchInfo, const std::string& stylesheet = "base");
	virtual ~RenderObjects();
	bool process(monad_m fm, monad_m lm);
	std::string getDocument();
	void clean();
	std::string getError() { return m_error_message; };
 protected:
	virtual void startDocument();
	void appendError(const std::string& error);
	void calculatePriorities();
	bool readObjects();
	void objects2Document();
	void startElements(const ListOfID_D& id_d_list, monad_m monad);
	void endElements(const ListOfID_D& id_d_list, monad_m monad);
	void startElement(MemObject *pObj, const std::string& OTN);
	void endElement(MemObject *pObj, const std::string& OTN);
	bool loadObjectsFromMonads(const std::string& OTN, const SetOfMonads& som, const JSONValue *pFeaturesToGet);
	bool setDefaults();
	void addToFMSandLMS(monad_m first, monad_m last, id_d_t id_d)
	{
		Monad2ID_DListMap::iterator itfm = m_fms.find(first);
		if (itfm == m_fms.end()) {
			//m_fms.insert(std::make_pair(first, ListOfID_D(1, id_d)));
			m_fms[first] = ListOfID_D(1, id_d);
		} else {
			itfm->second.push_back(id_d);
		}

		Monad2ID_DListMap::iterator itlm = m_lms.find(last);
		if (itlm == m_lms.end()) {
			//m_lms.insert(std::make_pair(last, ListOfID_D(1, id_d)));
			m_lms[last] = ListOfID_D(1, id_d);
		} else {
			itlm->second.push_back(id_d);
		}
	};
	const ListOfID_D& getLOIDDFromFMS(monad_m monad) 
	{
		Monad2ID_DListMap::iterator itfm = m_fms.find(monad);
		if (itfm == m_fms.end()) {
			static ListOfID_D empty;
			return empty;
		} else {
			return itfm->second;
		}
	};
	const ListOfID_D& getLOIDDFromLMS(monad_m monad)
	{
		Monad2ID_DListMap::iterator itlm = m_lms.find(monad);
		if (itlm == m_lms.end()) {
			static ListOfID_D empty;
			return empty;
		} else {
			return itlm->second;
		}
	}
};

#endif
