/*
 * string_func.h
 *
 * String helper functions
 *
 * Ulrik Petersen
 * Created: 3/1-2001
 * Last update: 5/25-2008
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2008  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file string_func.h
 *@brief String utility functions (EMdF layer)
 */


#ifndef STRING_FUNC__H__
#define STRING_FUNC__H__

#include <cstring>
#include <emdf.h>
#include <emdf_enums.h>
#include <string>
#include <cstdio>
#include <sstream>
#include <list>
#include <vector>

class IntegerList; // Forward declaration

extern void str_toupper(const std::string& in, std::string& result);
extern void str_tolower(const std::string& in, std::string& result);
extern int strcmp_nocase(const std::string& str1, const std::string& str2);
extern int strcmp_case(const std::string& str1, const std::string& str2);
extern std::string bool2string(bool b);
extern bool string2bool(const std::string& str);
extern std::string bool_alpha2string(bool b);
extern bool string2bool_alpha(const std::string& str);
extern std::string int2string(int i);
extern std::string id_d2string(id_d_t i);
extern std::string id_d2number_string(id_d_t i);
extern id_d_t string2id_d(const std::string& str);
extern std::string double2string(double d);
extern std::string monad_m2string(monad_m m);
extern monad_m string2monad_m(const std::string& str);
extern std::string long2string(long l);
extern long string2long(const std::string& str);
extern std::string encodeSTRINGstring(const std::string& str);
extern bool is_other_than_whitespace(const std::string& str);
inline bool is_whitespace(char c)
{
	return c == ' ' || c == '\n' || c == '\t' || c == '\r'; // NOTE: '\v' is missing, but I don't think it is ever used!
}

extern bool has_double_quote(const std::string& str);
extern bool is_hex(const std::string& str);
extern bool is_identifier(const std::string& str);
extern std::string convert_to_identifier(const std::string& instr);

extern void replace_chars(std::string &s, const char in, const char out);

// Turns _underscores into spaces
extern void replace_underscores(std::string& s);

// Turns spaces into _underscores
extern void replace_spaces(std::string& s);

// Replaces substrings with other strings
extern std::string replace_substring(const std::string& instring, std::string substring_to_replace, std::string replacement_string);

extern std::string reverse_string(const std::string& in);

// Checks whether string is an integer.
extern bool string_is_number(const std::string& in);

extern bool has_double_quote(const std::string& str);

extern std::string convert_to_hex(std::string instr);

extern long hex2long(const std::string& instr);

extern std::string remove_char(std::string instr, char char_to_remove);

extern std::string remove_from_substring_and_out(const std::string& instr, const std::string& substr);

extern std::string remove_chars_in_string(const std::string& instr, const std::string& chars_to_remove);

extern std::string remove_punct(const std::string instr);

bool string_ends_with(const std::string& input, const std::string& end);





/** Convert two characters representing a hex value to char.
 *
 * @param a The most significant nybble.
 *
 * @param b The least significant nybble.
 *
 * @return The parameters as a hex value.
 */
inline unsigned char hex2char(unsigned char a, unsigned char b)
{ // TODO: This could be faster... don't use sscanf, but do it directly! 
  char szTemp[5];
  sprintf(szTemp, "0x%c%c", a, b);
  unsigned int c;
  sscanf(szTemp, "%x", &c);
  return (unsigned char) c;
}

/** Convert a two-character string in HEX to char.
 * 
 * @param h The string to convert.
 *
 * @return The value of the hex-based string h.
 */
inline unsigned char hex2char(const char* h)
{
  return hex2char(h[0], h[1]);
}

/** Convert a char to a two-char hex value with lower-case letters.
 * 
 * @param c The char to convert.
 *
 * @return The string representation of the char c.
 */
inline std::string char2hex(unsigned char c)
{
  char szResult[5];
  sprintf(szResult, "%02x", c);
  return szResult;
}

/** Convert a char to a two-char hex value with upper-case letters.
 * 
 * @param c The char to convert.
 *
 * @return The string representation of the char c.
 */
inline std::string char2upperhex(unsigned char c)
{
  char szResult[5];
  sprintf(szResult, "%02X", c);
  return szResult;
}

/** Convert an octal string to a number (char).
 *
 * @param o The NUL-terminated string in octal to convert.
 *
 * @return The value of the octal-based string o.
 */
inline unsigned char octal2char(const char* o)
{
  unsigned int result;
  sscanf(o, "%o", &result);
  return (unsigned char) result;
}

/** Convert a character to 3-character octal octal.
 *
 * @param c The character to convert.
 *
 * @return The character as an octal string.
 */
inline std::string char2octal(unsigned char c)
{
  char szResult[6];
  sprintf(szResult, "%03o", c);
  return szResult;
}


bool string2charset(const std::string& input, eCharsets& charset);
bool charset2string(eCharsets charset, std::string& output);

bool string2backend_kind(const std::string& input, eBackendKind& backend_kind);
std::string backend_kind2string(eBackendKind backend_kind);

void long2utf8(long input, std::string& output);


/** A template to join a list of T with the string between inbetween
 *  each list member.
 */
template<class T> std::string joinList(std::string between, const std::list<T>& l)
{
	std::ostringstream str;
	typename std::list<T>::const_iterator ci = l.begin();
	if (ci != l.end()) {
		str << *ci;
		++ci;
	}
	while (ci != l.end()) {
		str << between << *ci;
		++ci;
	}
	return str.str();
}

/** A template to join a list of T into one long string
 */
template<class T> std::string joinListEmptyBetween(const std::list<T>& l)
{
	std::ostringstream str;
	typename std::list<T>::const_iterator ci = l.begin();
	typename std::list<T>::const_iterator cend = l.end();
	while (ci != cend) {
		str << *ci;
		++ci;
	}
	return str.str();
}






extern void split_string(const std::string& instring, const std::string& splitchars, std::list<std::string>& outlist);

extern void split_string(const std::string& instring, const std::string& splitchars, std::vector<std::string>& outlist);

extern void split_string_on_string(const std::string& instring, const std::string& splitstring, std::list<std::string>& outlist);

extern void split_string_on_string(const std::string& instring, const std::string& splitstring, std::vector<std::string>& outlist);

extern std::string parse_C_backslashes(const std::string& instring);

extern std::string strip(const std::string& instring, std::string stripchars);

// strip " \n\r\v\t"
extern std::string strip(const std::string& instring);


extern std::string long2compact_string7bits(long input);
extern void compact_string7bits2long(const std::string& input, long& output);
extern void IntegerList2compact_string7bits(const IntegerList& intlist, std::string& result);
extern void compact_string7bits2IntegerList(const std::string& inStr, IntegerList& intlist);

inline std::string escapeXMLEntities(const std::string& input)
{
	std::string result = replace_substring(input, "&", "&amp;");
	result = replace_substring(result, "\"", "&quot;");
	result = replace_substring(result, "<", "&lt;");
	result = replace_substring(result, ">", "&gt;");
	return result;
}





#endif // STRING_FUNC__H__

